import shutil
import pathlib
import os


from eu_consultations.scrape import scrape, read_initiatives_from_json, scrape_topics
from eu_consultations.extract_filetext import (
    download_consultation_files,
    extract_text_from_attachments,
)
from eu_consultations.consultation_data import Consultation, Initiative

OUTPUT_FILENAME = "consultation_data.json"


def test_scraping_creates_output_file(test_output_dir):
    scrape(
        topic_list=["DIGITAL"],
        output_folder=test_output_dir,
        filename=OUTPUT_FILENAME,
        max_pages=1,
    )
    # list all files
    out = pathlib.Path(test_output_dir)
    file_list = [str(item.name) for item in list(out.rglob("*")) if item.is_file()]
    shutil.rmtree(test_output_dir)
    assert OUTPUT_FILENAME in file_list


def test_scraped_data_conforms_to_dataclass(test_output_dir):
    initiatives = scrape(
        topic_list=["DIGITAL"],
        output_folder=test_output_dir,
        filename=OUTPUT_FILENAME,
        max_pages=1,
    )
    stored_consultations = read_initiatives_from_json(
        os.path.join(test_output_dir, OUTPUT_FILENAME)
    )
    shutil.rmtree(test_output_dir)
    assert all([isinstance(item, Initiative) for item in initiatives])
    assert all([isinstance(item, Initiative) for item in stored_consultations])


def test_full_pipeline_runs(test_output_dir):
    scrape_folder = test_output_dir
    # get minimal consultation data --------
    initiatives_data = scrape_topics(
        topic_list=["DIGITAL"],
        max_pages=1,
        max_feedback=5,
    )
    # download files and add download path to data --------
    output_folder = os.path.join(scrape_folder, "files")
    data_with_downloads = download_consultation_files(
        initiatives_data=initiatives_data, output_folder=output_folder
    )
    data_with_extracted_text = extract_text_from_attachments(
        initiatives_data_with_attachments=data_with_downloads,
        stream_out_folder=output_folder,
    )
    # capture consultation ids
    consultation_ids = [c.id for c in initiatives_data]
    after_download_consultation_ids = [c.id for c in data_with_downloads]
    after_extraction_consultation_ids = [c.id for c in data_with_extracted_text]
    # remove output folder
    shutil.rmtree(test_output_dir)
    # basic data integrity tests
    assert (
        set(consultation_ids)
        == set(after_download_consultation_ids)
        == set(after_extraction_consultation_ids)
    )
    assert all([isinstance(item, Initiative) for item in data_with_downloads])
    assert all([isinstance(item, Initiative) for item in data_with_extracted_text])



import os
from typing import List

import httpx
from docling.document_converter import DocumentConverter
import srsly
from loguru import logger
from tqdm import tqdm

from eu_consultations.consultation_data import Initiative, DoclingDump
from eu_consultations.scrape import save_to_json


def download_file(url, filepath: str | os.PathLike) -> None | str | os.PathLike:
    """download file from url and return filepath of download location

    Args:
        url: url to download
        filepath: download filepath
    Returns:
        String with filepath to download location or None
    """
    try:
        r = httpx.get(url, timeout=None)
        with open(filepath, "wb") as f:
            f.write(r.content)
        logger.info(f"Downloaded {url}")
        return filepath
    except httpx.HTTPError as exc:
        logger.error(f"Could not download {exc.request.url} - {exc}.")


def download_consultation_files(
    initiatives_data: List[Initiative], output_folder: str | os.PathLike
) -> List[Initiative]:
    """download all files referenced in consultation data and return data with path to files appended

    Args:
        initiatives_data: List of Initiative dataclass objects
        output_folder: Path to folder to store files in

    Returns:
        List of Initiative dataclass objects with downloaded_filepath attribute set on attachments
    """
    # process all pdfs in a scrape
    if not os.path.exists(output_folder):
        os.makedirs(output_folder)
    for initiative in initiatives_data:
        if initiative.consultations is not None:
            for consultation in tqdm(initiative.consultations, desc="Consultations"):
                if consultation.feedback is not None:
                    for feedback in tqdm(consultation.feedback, desc="Feedback"):
                        if feedback:
                            for attachment in feedback.attachments:
                                downloaded_filepath = download_file(
                                    url=attachment.links,
                                    filepath=os.path.join(
                                        output_folder,
                                        f"{attachment.document_id}_{attachment.file_name}",
                                    ),
                                )
                                if downloaded_filepath is not None:
                                    attachment.downloaded_filepath = downloaded_filepath
    return initiatives_data


def extract_text_from_attachments(
    initiatives_data_with_attachments: List[Initiative],
    stream_out_folder: str | os.PathLike | None,
    converter: DocumentConverter = DocumentConverter(),
) -> List[Initiative]:
    """extract text from all files referenced in consultation data using docling and return data with text appended

    Args:
        initiatives_data_with_attachments: List of Initiative dataclass objects with references to downloaded files set (probably created with download_consultation_files)
        stream_out_folder: Path to folder to stream out extracted texts as lossless JSON in Docling format per document (in subfolder "docling") and per consultation (in subfolder "consultation"). Defaults to none (no streaming out).
        converter: supply non-default DocumentConverter from docling

    Returns:
        List of Consultation dataclass objects with extracted text and docling json set on attachments
    """
    for initiative in initiatives_data_with_attachments:
        if initiative.consultations is not None:
            for consultation in tqdm(initiative.consultations, desc="Consultations"):
                if consultation.feedback is not None:
                    for feedback in tqdm(
                        consultation.feedback,
                        desc=f"Feedback for consultation {consultation.id} in inititative {initiative.short_title}",
                    ):
                        if feedback.attachments is not None:
                            for attachment in feedback.attachments:
                                logger.info(f"Extracting {attachment.downloaded_filepath}")
                                result = converter.convert(attachment.downloaded_filepath)
                                attachment.extracted_text = result.document.export_to_text()
                                docling_dict = result.document.export_to_dict()
                                docling_dump = DoclingDump.from_dict(docling_dict)
                                attachment.docling_json = docling_dump
                                if stream_out_folder is not None:
                                    docling_out_folder = os.path.join(stream_out_folder, "docling")
                                    if not os.path.exists(docling_out_folder):
                                        os.makedirs(docling_out_folder)
                                    srsly.write_json(
                                        os.path.join(
                                            docling_out_folder,
                                            f"{str(attachment.document_id)}.json",
                                        ),
                                        docling_dict,
                                    )
                    if stream_out_folder is not None:
                        save_to_json(
                            initiatives=[initiative],
                            output_folder=os.path.join(stream_out_folder, "consultations"),
                            filename=f"{str(initiative.id)}.json",
                        )
    return initiatives_data_with_attachments



from typing import List

import httpx
from tqdm import tqdm
from loguru import logger

from eu_consultations.consultation_data import Initiative, Consultation, Feedback

TOPICS = {
    "AGRI": "Agriculture and rural development",
    "FINANCE": "Banking and financial services",
    "BORDERS": "Borders and security",
    "BUDGET": "Budget",
    "BUSINESS": "Business and industry",
    "CLIMA": "Climate action",
    "COMP": "Competition",
    "CONSUM": "Consumers",
    "CULT": "Culture and media",
    "CUSTOMS": "Customs",
    "DIGITAL": "Digital economy and society",
    "ECFIN": "Economy, finance and the euro",
    "EAC": "Education and training",
    "EMPL": "Employment and social affairs",
    "ENER": "Energy",
    "ENV": "Environment",
    "ENLARG": "EU enlargement",
    "NEIGHBOUR": "European neighbourhood policy",
    "FOOD": "Food safety",
    "FOREIGN": "Foreign affairs and security policy",
    "FRAUD": "Fraud prevention",
    "HOME": "Home affairs",
    "HUMAN": "Humanitarian aid and civil protection",
    "INST": "Institutional affairs",
    "INTDEV": "International cooperation and development",
    "JUST": "Justice and fundamental rights",
    "MARE": "Maritime affairs and fisheries",
    "ASYL": "Migration and asylum",
    "HEALTH": "Public health",
    "REGIO": "Regional policy",
    "RESEARCH": "Research and innovation",
    "SINGMARK": "Single market",
    "SPORT": "Sport",
    "STAT": "Statistics",
    "TAX": "Taxation",
    "TRADE": "Trade",
    "TRANSPORT": "Transport",
    "YOUTH": "Youth",
}

INITIATIVES_PAGE_SIZE = 10

BASE_URL = "https://ec.europa.eu/info/law/better-regulation"

API_PATH_INITIATIVES = "/brpapi/searchInitiatives?"
API_PATH_PUBLICATIONS = "/brpapi/groupInitiatives/"
FEEDBACK_URL = "/api/allFeedback"


# getting consultation ids


def get_total_pages_initiatives(topic: str) -> int:
    url = BASE_URL + API_PATH_INITIATIVES
    params = {"topic": topic, "page": 0, "size": INITIATIVES_PAGE_SIZE}
    r = httpx.get(url, params=params, timeout=None)
    total_pages = r.json()["page"]["totalPages"]
    return total_pages


def get_ids_for_page_initiatives(topic, page) -> List[int]:
    url = BASE_URL + API_PATH_INITIATIVES
    params = {"topic": topic, "page": page, "size": INITIATIVES_PAGE_SIZE}
    r = httpx.get(url, params=params, timeout=None)
    initiatives = r.json()["_embedded"]["initiativeResultDtoes"]
    ids = [int(initiative["id"]) for initiative in initiatives]
    return ids


def get_ids_for_topic(topic: str, max_pages: int | None = None) -> List[int]:
    if max_pages is not None:
        total_pages = max_pages
        logger.warning(f"scrapping only {max_pages} page due to max_pages setting")
    else:
        total_pages = get_total_pages_initiatives(topic)
    ids = []
    for page in tqdm(range(0, total_pages), desc="Gathering page ids"):
        page_ids = get_ids_for_page_initiatives(topic, page)
        ids.extend(page_ids)
    return ids


# get data on initiatives for specific initiative id


def get_initiative_data(id: int) -> Initiative:
    # get all publications under id
    url = BASE_URL + API_PATH_PUBLICATIONS + str(id)
    r = httpx.get(url)
    data = r.json()
    consultations_list = Consultation.from_list(data["publications"])
    initiative = Initiative.from_dict(
        {i: data[i] for i in data if i not in "publications"}
    )
    initiative.consultations = consultations_list
    return initiative


# getting publications with potential feedbacks for initiatives


def get_total_pages_feedback(consultation: Consultation) -> int:
    url = BASE_URL + "/api/allFeedback"
    params = {
        "publicationId": consultation.id,
        "size": INITIATIVES_PAGE_SIZE,
        "page": 0,
    }
    r = httpx.get(url, params=params)
    total_pages = r.json()["page"]["totalPages"]
    return total_pages


def get_feedback_for_page(consultation, page) -> List[Feedback]:
    url = BASE_URL + "/api/allFeedback"
    params = {
        "publicationId": consultation.id,
        "size": INITIATIVES_PAGE_SIZE,
        "page": page,
    }
    r = httpx.get(url, params=params)
    # remove _links data
    feedback_data = r.json()["_embedded"]["feedback"]
    for feedback in feedback_data:
        if "_links" in feedback:
            del feedback["_links"]
            # Process attachments field
        if "attachments" in feedback:
            for attachment in feedback["attachments"]:
                attachment["links"] = (
                    f"https://ec.europa.eu/info/law/better-regulation/api/download/{attachment['documentId']}"
                )
                # Remove unwanted fields
                keys_to_remove = set(attachment.keys()) - {
                    "id",
                    "fileName",
                    "documentId",
                    "links",
                }
                for key in keys_to_remove:
                    del attachment[key]
    feedbacks = Feedback.from_list(feedback_data)
    return feedbacks


def get_feedback_for_consultation(
    consultation: Consultation, max_feedback: int | None = None
) -> List[Feedback] | None:
    if consultation.total_feedback == 0:
        return None
    if max_feedback is not None:
        if consultation.total_feedback > max_feedback:
            logger.warning(
                f"skipping consultation {consultation.id} because of max_feedback setting"
            )
            return None
    else:
        total_pages = get_total_pages_feedback(consultation=consultation)
        feedbacks = []
        for page in tqdm(
            range(0, total_pages),
            desc=f"Gathering feedback for consultation {consultation.id}",
        ):
            feedback_per_page = get_feedback_for_page(
                consultation=consultation, page=page
            )
            feedbacks.extend(feedback_per_page)
        return feedbacks


def scrape_topic(
    topic: str,
    max_pages: int | None = None,
    max_feedback: int | None = None,
) -> List[Initiative]:
    initiative_ids = get_ids_for_topic(topic, max_pages=max_pages)
    initiatives = []
    for id in tqdm(initiative_ids, desc="Gathering initiative data"):
        initiative = get_initiative_data(id)
        initiatives.append(initiative)
    for initiative in tqdm(initiatives, desc="Processing initiatives"):
        for consultation in initiative.consultations:
            try:
                feedbacks = get_feedback_for_consultation(
                    consultation, max_feedback=max_feedback
                )
                consultation.feedback = feedbacks
            except Exception as e:
                logger.error(
                    f"Could not process consultation {consultation.id} initiative {initiative.id}: {e}"
                )
    return initiatives
